/**
 * bd_fs_ext4_info_copy: (skip)
 *
 * Creates a new copy of @data.
 */
BDFSExt4Info* bd_fs_ext4_info_copy (BDFSExt4Info *data) {
        BDFSExt4Info *ret = g_new0 (BDFSExt4Info, 1);

    ret->label = g_strdup (data->label);
    ret->uuid = g_strdup (data->uuid);
    ret->state = g_strdup (data->state);
    ret->block_size = data->block_size;
    ret->block_count = data->block_count;
    ret->free_blocks = data->free_blocks;

    return ret;
}

/**
 * bd_fs_ext4_info_free: (skip)
 *
 * Frees @data.
 */
void  bd_fs_ext4_info_free (BDFSExt4Info *data) {
        g_free (data->label);
    g_free (data->uuid);
    g_free (data->state);
    g_free (data);
}

GType  bd_fs_ext4_info_get_type () {
        static GType type = 0;

    if (G_UNLIKELY(type == 0)) {
        type = g_boxed_type_register_static("BDFSExt4Info",
                                            (GBoxedCopyFunc) bd_fs_ext4_info_copy,
                                            (GBoxedFreeFunc) bd_fs_ext4_info_free);
    }

    return type;
}

/**
 * bd_fs_xfs_info_copy: (skip)
 *
 * Creates a new copy of @data.
 */
BDFSXfsInfo* bd_fs_xfs_info_copy (BDFSXfsInfo *data) {
        BDFSXfsInfo *ret = g_new0 (BDFSXfsInfo, 1);

    ret->label = g_strdup (data->label);
    ret->uuid = g_strdup (data->uuid);
    ret->block_size = data->block_size;
    ret->block_count = data->block_count;

    return ret;
}

/**
 * bd_fs_xfs_info_free: (skip)
 *
 * Frees @data.
 */
void  bd_fs_xfs_info_free (BDFSXfsInfo *data) {
        g_free (data->label);
    g_free (data->uuid);
    g_free (data);
}

GType  bd_fs_xfs_info_get_type () {
        static GType type = 0;

    if (G_UNLIKELY(type == 0)) {
        type = g_boxed_type_register_static("BDFSXfsInfo",
                                            (GBoxedCopyFunc) bd_fs_xfs_info_copy,
                                            (GBoxedFreeFunc) bd_fs_xfs_info_free);
    }

    return type;
}

static const gchar const * const fs_functions[] = {
    "bd_fs_wipe",
    "bd_fs_ext4_mkfs",
    "bd_fs_ext4_wipe",
    "bd_fs_ext4_check",
    "bd_fs_ext4_repair",
    "bd_fs_ext4_set_label",
    "bd_fs_ext4_get_info",
    "bd_fs_ext4_resize",
    "bd_fs_xfs_mkfs",
    "bd_fs_xfs_wipe",
    "bd_fs_xfs_check",
    "bd_fs_xfs_repair",
    "bd_fs_xfs_set_label",
    "bd_fs_xfs_get_info",
    "bd_fs_xfs_resize",
    NULL};

gchar const * const * get_fs_functions (void) {
    return fs_functions;
}

static const guint8 fs_num_functions = 15;

guint8 get_fs_num_functions (void) {
    return fs_num_functions;
}

gboolean  bd_fs_wipe_stub (gchar *device, gboolean all, GError **error) {
    g_critical ("The function 'bd_fs_wipe' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_wipe' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_wipe) (gchar *device, gboolean all, GError **error) = bd_fs_wipe_stub;

/**
 * bd_fs_wipe:
 * @device: the device to wipe signatures from
 * @all: whether to wipe all (%TRUE) signatures or just the first (%FALSE) one
 * @error: (out): place to store error (if any)
 *
 * Returns: whether signagures were successfully wiped on @device or not
 */
gboolean  bd_fs_wipe (gchar *device, gboolean all, GError **error) {
    return _bd_fs_wipe (device, all, error);
}


gboolean  bd_fs_ext4_mkfs_stub (gchar *device, GError **error) {
    g_critical ("The function 'bd_fs_ext4_mkfs' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_ext4_mkfs' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_ext4_mkfs) (gchar *device, GError **error) = bd_fs_ext4_mkfs_stub;

/**
 * bd_fs_ext4_mkfs:
 * @device: the device to create a new ext4 fs on
 * @error: (out): place to store error (if any)
 *
 * Returns: whether a new ext4 fs was successfully created on @device or not
 */
gboolean  bd_fs_ext4_mkfs (gchar *device, GError **error) {
    return _bd_fs_ext4_mkfs (device, error);
}


gboolean  bd_fs_ext4_wipe_stub (gchar *device, GError **error) {
    g_critical ("The function 'bd_fs_ext4_wipe' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_ext4_wipe' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_ext4_wipe) (gchar *device, GError **error) = bd_fs_ext4_wipe_stub;

/**
 * bd_fs_ext4_wipe:
 * @device: the device to wipe an ext4 signature from
 * @error: (out): place to store error (if any)
 *
 * Returns: whether an ext4 signature was successfully wiped from the @device or
 *          not
 */
gboolean  bd_fs_ext4_wipe (gchar *device, GError **error) {
    return _bd_fs_ext4_wipe (device, error);
}


gboolean  bd_fs_ext4_check_stub (gchar *device, GError **error) {
    g_critical ("The function 'bd_fs_ext4_check' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_ext4_check' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_ext4_check) (gchar *device, GError **error) = bd_fs_ext4_check_stub;

/**
 * bd_fs_ext4_check:
 * @device: the device the file system on which to check
 * @error: (out): place to store error (if any)
 *
 * Returns: whether an ext4 file system on the @device is clean or not
 */
gboolean  bd_fs_ext4_check (gchar *device, GError **error) {
    return _bd_fs_ext4_check (device, error);
}


gboolean  bd_fs_ext4_repair_stub (gchar *device, gboolean unsafe, GError **error) {
    g_critical ("The function 'bd_fs_ext4_repair' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_ext4_repair' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_ext4_repair) (gchar *device, gboolean unsafe, GError **error) = bd_fs_ext4_repair_stub;

/**
 * bd_fs_ext4_repair:
 * @device: the device the file system on which to repair
 * @unsafe: whether to do unsafe operations too
 * @error: (out): place to store error (if any)
 *
 * Returns: whether an ext4 file system on the @device was successfully repaired
 *          (if needed) or not (error is set in that case)
 */
gboolean  bd_fs_ext4_repair (gchar *device, gboolean unsafe, GError **error) {
    return _bd_fs_ext4_repair (device, unsafe, error);
}


gboolean  bd_fs_ext4_set_label_stub (gchar *device, gchar *label, GError **error) {
    g_critical ("The function 'bd_fs_ext4_set_label' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_ext4_set_label' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_ext4_set_label) (gchar *device, gchar *label, GError **error) = bd_fs_ext4_set_label_stub;

/**
 * bd_fs_ext4_set_label:
 * @device: the device the file system on which to set label for
 * @label: label to set
 * @error: (out): place to store error (if any)
 *
 * Returns: whether the label of ext4 file system on the @device was
 *          successfully set or not
 */
gboolean  bd_fs_ext4_set_label (gchar *device, gchar *label, GError **error) {
    return _bd_fs_ext4_set_label (device, label, error);
}


BDFSExt4Info* bd_fs_ext4_get_info_stub (gchar *device, GError **error) {
    g_critical ("The function 'bd_fs_ext4_get_info' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_ext4_get_info' called, but not implemented!");
    return NULL;
}

BDFSExt4Info* (*_bd_fs_ext4_get_info) (gchar *device, GError **error) = bd_fs_ext4_get_info_stub;

/**
 * bd_fs_ext4_get_info:
 * @device: the device the file system of which to get info for
 * @error: (out): place to store error (if any)
 *
 * Returns: (transfer full): information about the file system on @device or
 *                           %NULL in case of error
 */
BDFSExt4Info* bd_fs_ext4_get_info (gchar *device, GError **error) {
    return _bd_fs_ext4_get_info (device, error);
}


gboolean  bd_fs_ext4_resize_stub (gchar *device, guint64 new_size, GError **error) {
    g_critical ("The function 'bd_fs_ext4_resize' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_ext4_resize' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_ext4_resize) (gchar *device, guint64 new_size, GError **error) = bd_fs_ext4_resize_stub;

/**
 * bd_fs_ext4_resize:
 * @device: the device the file system of which to resize
 * @new_size: new requested size for the file system (if 0, the file system is
 *            adapted to the underlying block device)
 * @error: (out): place to store error (if any)
 *
 * Returns: whether the file system on @device was successfully resized or not
 */
gboolean  bd_fs_ext4_resize (gchar *device, guint64 new_size, GError **error) {
    return _bd_fs_ext4_resize (device, new_size, error);
}


gboolean  bd_fs_xfs_mkfs_stub (gchar *device, GError **error) {
    g_critical ("The function 'bd_fs_xfs_mkfs' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_xfs_mkfs' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_xfs_mkfs) (gchar *device, GError **error) = bd_fs_xfs_mkfs_stub;

/**
 * bd_fs_xfs_mkfs:
 * @device: the device to create a new xfs fs on
 * @error: (out): place to store error (if any)
 *
 * Returns: whether a new xfs fs was successfully created on @device or not
 */
gboolean  bd_fs_xfs_mkfs (gchar *device, GError **error) {
    return _bd_fs_xfs_mkfs (device, error);
}


gboolean  bd_fs_xfs_wipe_stub (gchar *device, GError **error) {
    g_critical ("The function 'bd_fs_xfs_wipe' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_xfs_wipe' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_xfs_wipe) (gchar *device, GError **error) = bd_fs_xfs_wipe_stub;

/**
 * bd_fs_xfs_wipe:
 * @device: the device to wipe an xfs signature from
 * @error: (out): place to store error (if any)
 *
 * Returns: whether an xfs signature was successfully wiped from the @device or
 *          not
 */
gboolean  bd_fs_xfs_wipe (gchar *device, GError **error) {
    return _bd_fs_xfs_wipe (device, error);
}


gboolean  bd_fs_xfs_check_stub (gchar *device, GError **error) {
    g_critical ("The function 'bd_fs_xfs_check' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_xfs_check' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_xfs_check) (gchar *device, GError **error) = bd_fs_xfs_check_stub;

/**
 * bd_fs_xfs_check:
 * @device: the device containing the file system to check
 * @error: (out): place to store error (if any)
 *
 * Returns: whether an xfs file system on the @device is clean or not
 *
 * Note: if the file system is mounted it may be reported as unclean even if
 *       everything is okay and there are just some pending/in-progress writes
 */
gboolean  bd_fs_xfs_check (gchar *device, GError **error) {
    return _bd_fs_xfs_check (device, error);
}


gboolean  bd_fs_xfs_repair_stub (gchar *device, GError **error) {
    g_critical ("The function 'bd_fs_xfs_repair' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_xfs_repair' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_xfs_repair) (gchar *device, GError **error) = bd_fs_xfs_repair_stub;

/**
 * bd_fs_xfs_repair:
 * @device: the device containing the file system to repair
 * @error: (out): place to store error (if any)
 *
 * Returns: whether an xfs file system on the @device was successfully repaired
 *          (if needed) or not (error is set in that case)
 */
gboolean  bd_fs_xfs_repair (gchar *device, GError **error) {
    return _bd_fs_xfs_repair (device, error);
}


gboolean  bd_fs_xfs_set_label_stub (gchar *device, gchar *label, GError **error) {
    g_critical ("The function 'bd_fs_xfs_set_label' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_xfs_set_label' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_xfs_set_label) (gchar *device, gchar *label, GError **error) = bd_fs_xfs_set_label_stub;

/**
 * bd_fs_xfs_set_label:
 * @device: the device containing the file system to set label for
 * @label: label to set
 * @error: (out): place to store error (if any)
 *
 * Returns: whether the label of xfs file system on the @device was
 *          successfully set or not
 */
gboolean  bd_fs_xfs_set_label (gchar *device, gchar *label, GError **error) {
    return _bd_fs_xfs_set_label (device, label, error);
}


BDFSXfsInfo* bd_fs_xfs_get_info_stub (gchar *device, GError **error) {
    g_critical ("The function 'bd_fs_xfs_get_info' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_xfs_get_info' called, but not implemented!");
    return NULL;
}

BDFSXfsInfo* (*_bd_fs_xfs_get_info) (gchar *device, GError **error) = bd_fs_xfs_get_info_stub;

/**
 * bd_fs_xfs_get_info:
 * @device: the device containing the file system to get info for
 * @error: (out): place to store error (if any)
 *
 * Returns: (transfer full): information about the file system on @device or
 *                           %NULL in case of error
 */
BDFSXfsInfo* bd_fs_xfs_get_info (gchar *device, GError **error) {
    return _bd_fs_xfs_get_info (device, error);
}


gboolean  bd_fs_xfs_resize_stub (gchar *mpoint, guint64 new_size, GError **error) {
    g_critical ("The function 'bd_fs_xfs_resize' called, but not implemented!");
    g_set_error (error, BD_INIT_ERROR, BD_INIT_ERROR_NOT_IMPLEMENTED,
                "The function 'bd_fs_xfs_resize' called, but not implemented!");
    return FALSE;
}

gboolean  (*_bd_fs_xfs_resize) (gchar *mpoint, guint64 new_size, GError **error) = bd_fs_xfs_resize_stub;

/**
 * bd_fs_xfs_resize:
 * @mpoint: the mount point of the file system to resize
 * @new_size: new requested size for the file system *in file system blocks* (see bd_fs_xfs_get_info())
 *            (if 0, the file system is adapted to the underlying block device)
 * @error: (out): place to store error (if any)
 *
 * Returns: whether the file system mounted on @mpoint was successfully resized or not
 */
gboolean  bd_fs_xfs_resize (gchar *mpoint, guint64 new_size, GError **error) {
    return _bd_fs_xfs_resize (mpoint, new_size, error);
}


gpointer load_fs_from_plugin(gchar *so_name) {
    void *handle = NULL;
    char *error = NULL;
    gboolean (*check_fn) (void) = NULL;
    gboolean (*init_fn) (void) = NULL;

    handle = dlopen(so_name, RTLD_LAZY);
    if (!handle) {
        g_warning("failed to load module fs: %s", dlerror());
        return NULL;
    }

    dlerror();
    * (void**) (&check_fn) = dlsym(handle, "check");
    if ((error = dlerror()) != NULL)
        g_debug("failed to load the check() function for fs: %s", error);
    if (check_fn && !check_fn()) {
        dlclose(handle);
        return NULL;
    }    check_fn = NULL;

    dlerror();
    * (void**) (&init_fn) = dlsym(handle, "init");
    if ((error = dlerror()) != NULL)
        g_debug("failed to load the init() function for fs: %s", error);
    if (init_fn && !init_fn()) {
        dlclose(handle);
        return NULL;
    }    init_fn = NULL;

    dlerror();
    * (void**) (&_bd_fs_wipe) = dlsym(handle, "bd_fs_wipe");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_wipe: %s", error);

    dlerror();
    * (void**) (&_bd_fs_ext4_mkfs) = dlsym(handle, "bd_fs_ext4_mkfs");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_ext4_mkfs: %s", error);

    dlerror();
    * (void**) (&_bd_fs_ext4_wipe) = dlsym(handle, "bd_fs_ext4_wipe");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_ext4_wipe: %s", error);

    dlerror();
    * (void**) (&_bd_fs_ext4_check) = dlsym(handle, "bd_fs_ext4_check");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_ext4_check: %s", error);

    dlerror();
    * (void**) (&_bd_fs_ext4_repair) = dlsym(handle, "bd_fs_ext4_repair");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_ext4_repair: %s", error);

    dlerror();
    * (void**) (&_bd_fs_ext4_set_label) = dlsym(handle, "bd_fs_ext4_set_label");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_ext4_set_label: %s", error);

    dlerror();
    * (void**) (&_bd_fs_ext4_get_info) = dlsym(handle, "bd_fs_ext4_get_info");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_ext4_get_info: %s", error);

    dlerror();
    * (void**) (&_bd_fs_ext4_resize) = dlsym(handle, "bd_fs_ext4_resize");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_ext4_resize: %s", error);

    dlerror();
    * (void**) (&_bd_fs_xfs_mkfs) = dlsym(handle, "bd_fs_xfs_mkfs");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_xfs_mkfs: %s", error);

    dlerror();
    * (void**) (&_bd_fs_xfs_wipe) = dlsym(handle, "bd_fs_xfs_wipe");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_xfs_wipe: %s", error);

    dlerror();
    * (void**) (&_bd_fs_xfs_check) = dlsym(handle, "bd_fs_xfs_check");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_xfs_check: %s", error);

    dlerror();
    * (void**) (&_bd_fs_xfs_repair) = dlsym(handle, "bd_fs_xfs_repair");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_xfs_repair: %s", error);

    dlerror();
    * (void**) (&_bd_fs_xfs_set_label) = dlsym(handle, "bd_fs_xfs_set_label");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_xfs_set_label: %s", error);

    dlerror();
    * (void**) (&_bd_fs_xfs_get_info) = dlsym(handle, "bd_fs_xfs_get_info");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_xfs_get_info: %s", error);

    dlerror();
    * (void**) (&_bd_fs_xfs_resize) = dlsym(handle, "bd_fs_xfs_resize");
    if ((error = dlerror()) != NULL)
        g_warning("failed to load bd_fs_xfs_resize: %s", error);

    return handle;
}

gboolean unload_fs (gpointer handle) {
    _bd_fs_wipe = bd_fs_wipe_stub;
    _bd_fs_ext4_mkfs = bd_fs_ext4_mkfs_stub;
    _bd_fs_ext4_wipe = bd_fs_ext4_wipe_stub;
    _bd_fs_ext4_check = bd_fs_ext4_check_stub;
    _bd_fs_ext4_repair = bd_fs_ext4_repair_stub;
    _bd_fs_ext4_set_label = bd_fs_ext4_set_label_stub;
    _bd_fs_ext4_get_info = bd_fs_ext4_get_info_stub;
    _bd_fs_ext4_resize = bd_fs_ext4_resize_stub;
    _bd_fs_xfs_mkfs = bd_fs_xfs_mkfs_stub;
    _bd_fs_xfs_wipe = bd_fs_xfs_wipe_stub;
    _bd_fs_xfs_check = bd_fs_xfs_check_stub;
    _bd_fs_xfs_repair = bd_fs_xfs_repair_stub;
    _bd_fs_xfs_set_label = bd_fs_xfs_set_label_stub;
    _bd_fs_xfs_get_info = bd_fs_xfs_get_info_stub;
    _bd_fs_xfs_resize = bd_fs_xfs_resize_stub;

    return dlclose(handle) == 0;
}

